from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_READY


class ULTimeData(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.network = None
        self.TIME_PER_FRAME = ULOutSocket(
            self,
            self.get_time_per_frame
        )
        self.FPS = ULOutSocket(self, self.get_fps)
        self.TIMELINE = ULOutSocket(self, self.get_timeline)

    def get_time_per_frame(self):
        return self.network.time_per_frame

    def get_fps(self):
        tpf = self.network.time_per_frame
        if not tpf:
            return 1
        fps = (1 / tpf)
        return fps

    def get_timeline(self):
        return self.network.timeline

    def setup(self, network):
        self.network = network

    def has_status(self, status):
        return status is STATUS_READY

    def evaluate(self):
        pass



from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid
import math


class ULAbsoluteValue(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.out_value = False
        self.OUT = ULOutSocket(self, self._get_out_value)

    def _get_out_value(self):
        value = self.get_input(self.value)
        if is_invalid(self.value):
            return STATUS_WAITING
        return math.fabs(value)

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import debug
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting


class ULControllerStatus(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.obj_name = None
        self.cont_name = None
        self.OUT = ULOutSocket(self, self.get_controller)
        self.SENSORS = ULOutSocket(self, self.get_sensors)

    def get_controller(self):
        socket = self.get_output('controller')
        if socket is None:
            game_obj = self.game_obj
            cont = game_obj.controllers[self.controller]
            state = (
                game_obj
                .blenderObject
                .game
                .controllers[self.controller]
                .type
            )
            if not cont.sensors:
                return self.set_output(
                    'controller',
                    False
                )
            elif state == 'LOGIC_AND':
                return self.set_output(
                    'controller',
                    False not in [sens.positive for sens in cont.sensors]
                )
            elif state == 'LOGIC_OR':
                return self.set_output(
                    'controller',
                    True in [sens.positive for sens in cont.sensors]
                )
            elif state == 'LOGIC_NAND':
                return self.set_output(
                    'controller',
                    False in [sens.positive for sens in cont.sensors]
                )
            elif state == 'LOGIC_NOR':
                return self.set_output(
                    'controller',
                    True not in [sens.positive for sens in cont.sensors]
                )
            elif state == 'LOGIC_XOR':
                return self.set_output(
                    'controller',
                    [
                        sens.positive
                        for sens in
                        cont.sensors
                    ].count(True) % 2 != 0
                )
            elif state == 'LOGIC_XNOR':
                check = cont.sensors[0].positive
                return self.set_output(
                    'controller',
                    False not in [
                        sens.positive == check
                        for sens in
                        cont.sensors
                    ]
                )
            else:
                debug('Expression/Python not supported for controller.')
                return self.set_output(
                    'controller',
                    False
                )
        return socket

    def get_sensors(self):
        sensors = {}
        cont = self.controller
        for sens in self.game_obj.controllers[cont].sensors:
            sensors[sens] = sens.positive
        return sensors

    def evaluate(self):
        self.game_obj = self.get_input(self.obj_name)
        self.controller = self.get_input(self.cont_name)
        if is_waiting(self.controller):
            return
        if is_invalid(self.game_obj):
            return
        self._set_ready()



from bge import logic
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULMouseRayCast(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.distance = None
        self.property = None
        self.xray = None
        self.camera = None
        self._set_value(False)
        self._out_object = None
        self._out_normal = None
        self._out_point = None
        self.OUTOBJECT = ULOutSocket(self, self.get_out_object)
        self.OUTNORMAL = ULOutSocket(self, self.get_out_normal)
        self.OUTPOINT = ULOutSocket(self, self.get_out_point)

    def get_out_object(self):
        return self._out_object

    def get_out_normal(self):
        return self._out_normal

    def get_out_point(self):
        return self._out_point

    def evaluate(self):
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        distance = self.get_input(self.distance)
        property_name = self.get_input(self.property)
        xray = self.get_input(self.xray)
        camera = self.get_input(self.camera)
        if is_waiting(distance, property_name, xray, camera):
            return
        self._set_ready()
        if not condition:
            self._set_value(False)
            self._out_normal = None
            self._out_object = None
            self._out_point = None
            return
        if is_invalid(camera):
            return
        mpos = logic.mouse.position
        vec = 10 * camera.getScreenVect(*mpos)
        ray_target = camera.worldPosition - vec
        target, point, normal = self.network.ray_cast(
            camera,
            None,
            ray_target,
            property_name,
            xray,
            distance
        )
        self._set_value(target is not None)
        self._out_object = target
        self._out_normal = normal
        self._out_point = point



from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULCameraRayCast(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.camera = None
        self.aim = None
        self.property_name = None
        self.xray = None
        self.distance = None
        self._picked_object = None
        self._picked_point = None
        self._picked_normal = None
        self.PICKED_OBJECT = ULOutSocket(self, self.get_picked_object)
        self.PICKED_POINT = ULOutSocket(self, self.get_picked_point)
        self.PICKED_NORMAL = ULOutSocket(self, self.get_picked_normal)

    def get_picked_object(self):
        return self._picked_object

    def get_picked_point(self):
        return self._picked_point

    def get_picked_normal(self):
        return self._picked_normal

    def evaluate(self):
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        camera = self.get_input(self.camera)
        aim = self.get_input(self.aim)
        property_name = self.get_input(self.property_name)
        xray = self.get_input(self.xray)
        distance = self.get_input(self.distance)
        if is_waiting(camera, aim, property_name, xray, distance):
            return
        self._set_ready()
        if not condition:
            self._set_value(False)
            self._out_normal = None
            self._out_object = None
            self._out_point = None
            return
        if is_invalid(camera):
            return
        if is_invalid(aim):
            return
        obj, point, normal = None, None, None
        # assume screen coordinates
        if isinstance(aim, Vector) and len(aim) == 2:
            vec = 10 * camera.getScreenVect(aim[0], aim[1])
            ray_target = camera.worldPosition - vec
            aim = ray_target
        if not property_name:
            obj, point, normal = camera.rayCast(aim, None, distance)
        else:
            obj, point, normal = camera.rayCast(
                aim,
                None,
                distance,
                property_name,
                xray=xray
            )
        self._set_value(obj is not None)
        self._picked_object = obj
        self._picked_point = point
        self._picked_normal = normal



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.physics import RWD
from uplogic.utils import VEHICLE
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULVehicleApplyBraking(ULActionNode):
    def __init__(self, value_type=RWD):
        ULActionNode.__init__(self)
        self.value_type = str(value_type)
        self.condition = None
        self.vehicle = None
        self.wheelcount = None
        self._reset = False
        self.power = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        game_object = self.get_input(self.vehicle)
        if is_invalid(game_object):
            return
        vehicle = game_object.get(VEHICLE, None)
        if vehicle is None:
            return
        value_type = self.get_input(self.value_type)
        wheelcount = self.get_input(self.wheelcount)
        power = self.get_input(self.power)
        if is_waiting(value_type, wheelcount, power):
            return
        self._set_ready()
        vehicle.brake(power, value_type, wheelcount)
        self.done = True



from mathutils import Vector
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid
import math


class ULVectorAngle(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.op = None
        self.vector: Vector = None
        self.vector_2: Vector = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        vector: Vector = self.get_input(self.vector)
        vector_2: Vector = self.get_input(self.vector_2)
        if is_invalid(
            vector,
            vector_2
        ):
            return STATUS_WAITING
        rad: float = vector.angle(vector_2)
        deg: float = rad * 180/math.pi
        return deg

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULSetCurvePoints(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.curve_object = None
        self.points: list = None
        self.done: bool = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        curve_object = self.get_input(self.curve_object)
        points = self.get_input(self.points)
        if is_waiting(points):
            return
        if is_invalid(curve_object):
            return
        self._set_ready()
        if not points:
            return
        curve = curve_object.blenderObject.data
        for spline in curve.splines:
            curve.splines.remove(spline)
        spline = curve.splines.new('NURBS')
        pos = curve_object.worldPosition
        spline.points.add(len(points))
        for p, new_co in zip(spline.points, points):
            p.co = ([
                new_co.x - pos.x,
                new_co.y - pos.y,
                new_co.z - pos.z
            ] + [1.0])
        self.done = True



from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting


class ULValueChangedTo(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.monitored_value = None
        self.trigger_value = None
        self.changed = None
        self.old = None
        self.OUT = ULOutSocket(self, self.get_changed)

    def get_changed(self):
        socket = self.get_output('changed')
        if socket is None:
            monitored_value = self.get_input(self.monitored_value)
            trigger_value = self.get_input(self.trigger_value)
            if is_waiting(monitored_value, trigger_value):
                return STATUS_WAITING
            return self.set_output(
                'changed',
                (
                    monitored_value == trigger_value and
                    self.changed
                )
            )
        return socket

    def evaluate(self):
        monitored_value = self.get_input(self.monitored_value)
        self.changed = monitored_value != self.old
        self.old = monitored_value
        self._set_ready()



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import make_unique_light
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULMakeUniqueLight(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.light = None
        self.done = None
        self._light = None
        self.OUT = ULOutSocket(self, self.get_done)
        self.LIGHT = ULOutSocket(self, self.get_light)

    def get_done(self):
        return self.done

    def get_light(self):
        return self._light

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_value(False)
            return self._set_ready()
        old_lamp_ge = self.get_input(self.light)
        if is_waiting(old_lamp_ge):
            return
        self._set_ready()

        make_unique_light(old_lamp_ge)

        self.done = True



from .handleevent import ULHandleEvent  # noqa
from .checkdistance import ULCheckDistance  # noqa
from .collision import ULCollision  # noqa
from .compare import ULCompare  # noqa
from .comparevectors import ULCompareVectors  # noqa
from .evaluateproperty import ULEvaluateProperty  # noqa
from .gamepadactive import ULGamepadActive  # noqa
from .gamepadbutton import ULGamepadButton  # noqa
from .gamepadbuttonup import ULGamepadButtonUp  # noqa
from .hasproperty import ULHasProperty  # noqa
from .keyboardactive import ULKeyboardActive  # noqa
from .keypressed import ULKeyPressed  # noqa
from .keyreleased import ULKeyReleased  # noqa
from .logicand import ULAnd  # noqa
from .logicand import ULAndList  # noqa
from .logicandnot import ULAndNot  # noqa
from .logicnone import ULNone  # noqa
from .logicnot import ULNot  # noqa
from .logicnotnone import ULNotNone  # noqa
from .logicor import ULOr  # noqa
from .logicor import ULOrList  # noqa
from .logictreestatus import ULLogicTreeStatus  # noqa
from .mousemoved import ULMouseMoved  # noqa
from .mouseover import ULMouseOver  # noqa
from .mousepressed import ULMousePressed  # noqa
from .mousepressedon import ULMousePressedOn  # noqa
from .mousereleased import ULMouseReleased  # noqa
from .mousescroll import ULMouseScrolled  # noqa
from .once import ULOnce  # noqa
from .oninit import ULOnInit  # noqa
from .onnexttick import ULOnNextTick  # noqa
from .onupdate import ULOnUpdate  # noqa
from .onvaluechanged import ULOnValueChanged  # noqa
from .onvaluechangedto import ULValueChangedTo  # noqa
from .sensorpositive import ULSensorPositive  # noqa
from .timer import ULTimer  # noqa
from .truefalse import ULTrueFalse  # noqa
from .valuevalid import ULValueValid  # noqa

from .controllerstatus import ULControllerStatus  # noqa
from .pulsify import ULPulsify  # noqa
from .barrier import ULBarrier
from .delay import ULDelay



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met
import bpy


class ULSetNodeSocket(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.tree_name = None
        self.node_name = None
        self.input_slot = None
        self.value = None
        self.done = False
        self.OUT = ULOutSocket(self, self._get_done)

    def _get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        tree_name = self.get_input(self.tree_name)
        node_name = self.get_input(self.node_name)
        input_slot = self.get_input(self.input_slot)
        value = self.get_input(self.value)
        if is_waiting(node_name, input_slot, value):
            return
        if is_invalid(tree_name):
            return
        if condition:
            self.done = True
            self._set_ready()
            (
                bpy.data.node_groups[tree_name]
                .nodes[node_name]
                .inputs[input_slot]
                .default_value
            ) = value



from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket


class ULOnValueChanged(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.old = None
        self.new = None
        self.current_value = None
        self.initialize = False
        self.OUT = ULOutSocket(self, self.get_changed)
        self.OLD = ULOutSocket(
            self,
            self.get_previous_value
        )
        self.NEW = ULOutSocket(self, self.get_current_value)

    def get_changed(self):
        socket = self.get_output('changed')
        if socket is None:
            curr = self.get_input(self.current_value)
            if not self.initialize:
                self.initialize = True
                self.old = self.new = curr
                return self.set_output('changed', False)
            elif self.old != curr:
                self.new = curr
                return self.set_output('changed', True)
        return socket

    def get_previous_value(self):
        return self.old

    def get_current_value(self):
        return self.new

    def reset(self):
        super().reset()
        self.old = self.new

    def evaluate(self):
        self._set_ready()



from mathutils import Vector
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_READY


class ULMouseData(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.MX = ULOutSocket(self, self.getmx)
        self.MY = ULOutSocket(self, self.getmy)
        self.MDX = ULOutSocket(self, self.getmdx)
        self.MDY = ULOutSocket(self, self.getmdy)
        self.MDWHEEL = ULOutSocket(self, self.getmdwheel)
        self.MXY0 = ULOutSocket(self, self.getmxyz)
        self.MDXY0 = ULOutSocket(self, self.getmdxyz)

    def getmx(self):
        return self.network._last_mouse_position[0]

    def getmy(self):
        return self.network._last_mouse_position[1]

    def getmdx(self):
        return self.network.mouse_motion_delta[0]

    def getmdy(self):
        return self.network.mouse_motion_delta[1]

    def getmdwheel(self):
        return self.network.mouse_wheel_delta

    def getmxyz(self):
        mp = self.network._last_mouse_position
        return Vector((mp[0], mp[1], 0))

    def getmdxyz(self):
        mp = self.network.mouse_motion_delta
        return Vector((mp[0], mp[1], 0))

    def evaluate(self):
        self._set_ready()

    def has_status(self, status):
        return status is STATUS_READY



from uplogic.nodes import ULConditionNode
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULDelay(ULConditionNode):
    consumed: bool
    triggers: list

    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition = None
        self.delay = None
        self.triggers = []

    def evaluate(self):
        condition = self.get_input(self.condition)
        delay = self.get_input(self.delay)
        if is_invalid(delay):
            return
        self._set_ready()

        now = self.network.timeline

        if not not_met(condition):
            self.triggers.append(now + delay)

        if not self.triggers:
            self._set_value(False)
            return
        t = self.triggers[0]
        if now >= t:
            self._set_value(True)
            self.triggers.remove(t)
            return
        self._set_value(False)



from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import not_met
from uplogic.utils import rot_to
from uplogic.utils import move_to


class ULFollowPath(ULActionNode):
    class MotionPath(object):
        def __init__(self):
            self.points = []
            self.cursor = 0
            self.loop = False
            self.loop_start = 0

        def next_point(self):
            if self.cursor < len(self.points):
                return self.points[self.cursor]
            else:
                return None

        def advance(self):
            self.cursor += 1
            if self.cursor < len(self.points):
                return True
            else:
                if self.loop:
                    self.cursor = self.loop_start
                    return True
                return False

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.moving_object = None
        self.rotating_object = None
        self.path_points = None
        self.loop = None
        self.path_continue = None
        self.navmesh_object = None
        self.move_dynamic = None
        self.linear_speed = None
        self.reach_threshold = None
        self.look_at = None
        self.rot_speed = None
        self.rot_axis = None
        self.front_axis = None
        self._motion_path = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        path_continue = self.get_input(self.path_continue)
        if not_met(condition):
            if not path_continue:
                self._motion_path = None
            self._set_ready()
            return
        moving_object = self.get_input(self.moving_object)
        rotating_object = self.get_input(self.rotating_object)
        path_points = self.get_input(self.path_points)
        navmesh_object = self.get_input(self.navmesh_object)
        move_dynamic = self.get_input(self.move_dynamic)
        linear_speed = self.get_input(self.linear_speed)
        reach_threshold = self.get_input(self.reach_threshold)
        look_at = self.get_input(self.look_at)
        rot_axis = self.get_input(self.rot_axis)
        front_axis = self.get_input(self.front_axis)
        rot_speed = self.get_input(self.rot_speed)
        loop = self.get_input(self.loop)
        if is_invalid(
            path_points,
            move_dynamic,
            linear_speed,
            reach_threshold,
            look_at,
            rot_axis,
            front_axis,
            loop
        ):
            return
        if is_invalid(rot_speed):
            rot_speed = 0
        if loop is None:
            return
        if is_invalid(moving_object):
            return
        if is_invalid(navmesh_object):
            navmesh_object = None
        self._set_ready()
        self._set_value(False)
        if (self._motion_path is None) or (self._motion_path.loop != loop):
            self.generate_path(
                moving_object.worldPosition,
                path_points,
                navmesh_object,
                loop
            )
        next_point = self._motion_path.next_point()
        if next_point:
            tpf = self.network.time_per_frame
            if look_at:
                rot_to(
                    rot_axis,
                    rotating_object,
                    next_point,
                    front_axis,
                    rot_speed,
                    tpf
                )
            reached = move_to(
                moving_object,
                next_point,
                linear_speed,
                tpf,
                move_dynamic,
                reach_threshold
            )
            if reached:
                has_more = self._motion_path.advance()
                if not has_more:
                    self._set_value(True)
                    self.done = True

    def generate_path(self, start_position, path_points, navmesh_object, loop):
        if not path_points:
            return self._motion_path.points.clear()
        path = ULFollowPath.MotionPath()
        path.loop = loop
        points = path.points
        self._motion_path = path
        if not navmesh_object:
            points.append(Vector(start_position))
            if loop:
                path.loop_start = 1
            for p in path_points:
                points.append(Vector(p))
        else:
            last = path_points[-1]
            mark_loop_position = loop
            for p in path_points:
                subpath = navmesh_object.findPath(
                    start_position,
                    Vector(p)
                )
                if p is last:
                    points.extend(subpath)
                else:
                    points.extend(subpath[:-1])
                if mark_loop_position:
                    path.loop_start = len(points)
                    mark_loop_position = False
                start_position = Vector(p)
            if loop:
                subpath = navmesh_object.findPath(
                    Vector(last),
                    Vector(path_points[0])
                )
                points.extend(subpath[1:])



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULAppendListItem(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.items: list = None
        self.val = None
        self.new_list: list = None
        self.done: bool = None
        self.OUT = ULOutSocket(self, self.get_done)
        self.LIST = ULOutSocket(self, self.get_list)

    def get_done(self):
        return self.done

    def get_list(self):
        return self.new_list

    def evaluate(self):
        self.done: bool = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        list_d: list = self.get_input(self.items)
        val = self.get_input(self.val)
        if is_waiting(list_d, val):
            return
        self._set_ready()
        list_d.append(val)
        self.new_list = list_d
        self.done = True



from bge import logic
from bge import events
from bge import render


MOUSE_EVENTS = logic.mouse.inputs
'''Reference to `bge.logic.mouse.inputs`
'''

LMB = events.LEFTMOUSE
RMB = events.RIGHTMOUSE
MMB = events.MIDDLEMOUSE


class ULMouseData():
    def __init__(self) -> None:
        self.position = get_mouse_position()
        self.movement = (0, 0)
        self.wheel = mouse_wheel()
        logic.getCurrentScene().pre_draw.append(self.update)

    def update(self):
        old_pos = self.position
        new_pos = get_mouse_position()
        self.movement = (
            new_pos[0] - old_pos[0],
            new_pos[1] - old_pos[1]
        )
        self.position = new_pos
        self.wheel = mouse_wheel()

    def destroy(self):
        logic.getCurrentScene().pre_draw.remove(self.update)


def set_mouse_position(x: int, y: int, absolute: bool = False):
    if absolute:
        render.setMousePosition(x, y)
        return
    render.setMousePosition(
        int(x * render.getWindowWidth()),
        int(y * render.getWindowHeight())
    )


def get_mouse_position(absolute: bool = False):
    pos = logic.mouse.position
    if absolute:
        return (
            int(pos[0] * render.getWindowWidth()),
            int(pos[1] * render.getWindowHeight())
        )
    return pos


def mouse_moved(tap: bool = False) -> bool:
    '''Detect mouse movement.

    :param tap: Only use the first consecutive `True` output

    :returns: boolean
    '''
    if tap:
        return (
            MOUSE_EVENTS[events.MOUSEX].activated or
            MOUSE_EVENTS[events.MOUSEY].activated
        )
    else:
        return (
            MOUSE_EVENTS[events.MOUSEX].active or
            MOUSE_EVENTS[events.MOUSEY].active or
            MOUSE_EVENTS[events.MOUSEX].activated or
            MOUSE_EVENTS[events.MOUSEY].activated
        )


def mouse_tap(button=events.LEFTMOUSE) -> bool:
    '''Detect mouse button tap.

    :param button: can be either `LMB`, `RMB` or `MMB` from `uplogic.input`

    :returns: boolean
    '''
    return (
        MOUSE_EVENTS[button].activated or
        MOUSE_EVENTS[button].activated
    )


def mouse_down(button=events.LEFTMOUSE) -> bool:
    '''Detect mouse button held down.

    :param button: can be either `LMB`, `RMB` or `MMB` from `uplogic.input`

    :returns: boolean
    '''
    return (
        MOUSE_EVENTS[button].active or
        MOUSE_EVENTS[button].activated or
        MOUSE_EVENTS[button].active or
        MOUSE_EVENTS[button].activated
    )


def mouse_up(button=events.LEFTMOUSE) -> bool:
    '''Detect mouse button released.

    :param button: can be either `LMB`, `RMB` or `MMB` from `uplogic.input`

    :returns: boolean
    '''
    return (
        MOUSE_EVENTS[button].released or
        MOUSE_EVENTS[button].released
    )


def mouse_wheel(tap: bool = False) -> int:
    '''Detect mouse wheel activity.

    :param tap: Only use the first consecutive `True` output

    :returns: -1 if wheel down, 0 if idle, 1 if wheel up
    '''
    if tap:
        return (
            MOUSE_EVENTS[events.WHEELUPMOUSE].activated -
            MOUSE_EVENTS[events.WHEELDOWNMOUSE].activated
        )
    else:
        return (
            (
                MOUSE_EVENTS[events.WHEELUPMOUSE].activated or
                MOUSE_EVENTS[events.WHEELUPMOUSE].active
            ) - (
                MOUSE_EVENTS[events.WHEELDOWNMOUSE].activated or
                MOUSE_EVENTS[events.WHEELDOWNMOUSE].active
            )
        )



from uplogic.nodes import ULConditionNode
from uplogic.utils import is_waiting


class ULPulsify(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition = None
        self.delay = None
        self._triggered = False
        self._triggered_time = None
        self._trigger_delay = None

    def evaluate(self):
        socket = self.get_output('out')
        if socket is None:
            if self._triggered is True:
                self._set_ready()
                delta = self.network.timeline - self._triggered_time
                if delta < self._trigger_delay:
                    self._set_value(False)
                    return self.set_output('out', False)
            condition = self.get_input(self.condition)
            delay = self.get_input(self.delay)
            if is_waiting(condition, delay):
                return self.set_output('out', False)
            self._set_ready()
            self._set_value(False)
            if delay is None:
                return self.set_output('out', False)
            if condition is None:
                return self.set_output('out', False)
            if condition:
                self._triggered = True
                self._triggered_time = self.network.timeline
                self._trigger_delay = delay
                self._set_value(True)
                return self.set_output('out', True)
        return socket



from bge import logic
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULPlayAction(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.action_name = None
        self.stop_anim = None
        self.frames = None
        self.start_frame = None
        self.end_frame = None
        self.layer = None
        self.priority = None
        self.play_mode = None
        self.layer_weight = None
        self.old_layer_weight = None
        self.speed = None
        self.old_speed = None
        self.blendin = None
        self.blend_mode = None
        self._started = False
        self._running = False
        self._finished = False
        self._frame = 0.0
        self._finish_notified = False
        self.STARTED = ULOutSocket(self, self._get_started)
        self.FINISHED = ULOutSocket(self, self._get_finished)
        self.RUNNING = ULOutSocket(self, self._get_running)
        self.FRAME = ULOutSocket(self, self._get_frame)

    def _get_started(self):
        return self._started

    def _get_finished(self):
        return self._finished

    def _get_running(self):
        return self._running

    def _get_frame(self):
        return self._frame

    def _reset_subvalues(self):
        self._started = False
        self._finished = False
        self._running = False
        self._frame = 0.0
        self._finish_notified = False

    def _notify_finished(self, obj, layer):
        if not self._finish_notified and self.stop_anim:
            self._finish_notified = True
            self._finished = True
            obj.stopAction(layer)
        else:
            self._finished = False

    def evaluate(self):
        condition = self.get_input(self.condition)
        game_object = self.get_input(self.game_object)
        action_name = self.get_input(self.action_name)
        start_frame = self.get_input(self.start_frame)
        end_frame = self.get_input(self.end_frame)
        layer = self.get_input(self.layer)
        priority = self.get_input(self.priority)
        play_mode = self.get_input(self.play_mode)
        layer_weight = self.get_input(self.layer_weight)
        speed = self.get_input(self.speed)
        blendin = self.get_input(self.blendin)
        blend_mode = self.get_input(self.blend_mode)
        if is_invalid(
            game_object,
            action_name,
            start_frame,
            end_frame,
            layer,
            priority,
            play_mode,
            layer_weight,
            speed,
            blendin,
            blend_mode
        ):
            return
        if play_mode > 2:
            if not_met(condition):
                self._notify_finished(game_object, layer)
                return
            else:
                play_mode -= 3
        if layer_weight <= 0:
            layer_weight = 0.0
        elif layer_weight >= 1:
            layer_weight = 1.0
        if speed <= 0:
            speed = 0.01
        self._set_ready()
        if is_invalid(game_object):  # can't play
            self._reset_subvalues()
        else:
            # Condition might be false and the animation running
            # because it was started in a previous evaluation
            playing_action = game_object.getActionName(layer)
            playing_frame = game_object.getActionFrame(layer)
            min_frame = start_frame
            max_frame = end_frame
            if end_frame < start_frame:
                min_frame = end_frame
                max_frame = max_frame
            if (
                (playing_action == action_name) and
                (playing_frame >= min_frame) and
                (playing_frame <= max_frame)
            ):
                self._started = False
                self._running = True
                is_near_end = False
                self._frame = playing_frame
                if (
                    layer_weight != self.old_layer_weight or
                    speed != self.old_speed
                ):
                    reset_frame = (
                        start_frame
                        if play_mode == logic.KX_ACTION_MODE_LOOP else
                        end_frame
                    )
                    next_frame = (
                        playing_frame + speed
                        if
                        playing_frame + speed <= end_frame
                        else
                        reset_frame
                    )
                    game_object.stopAction(layer)
                    game_object.playAction(
                        action_name,
                        start_frame,
                        end_frame,
                        layer=layer,
                        priority=priority,
                        blendin=blendin,
                        play_mode=play_mode,
                        speed=speed,
                        layer_weight=1 - layer_weight,
                        blend_mode=blend_mode)
                    game_object.setActionFrame(next_frame, layer)
                # TODO: the meaning of start-end depends
                # also on the action mode
                if end_frame > start_frame:  # play 0 to 100
                    is_near_end = (playing_frame >= (end_frame - 0.5))
                else:  # play 100 to 0
                    is_near_end = (playing_frame <= (end_frame + 0.5))
                if is_near_end:
                    self._notify_finished(game_object, layer)
            elif condition:  # start the animation if the condition is True
                is_playing = game_object.isPlayingAction(layer)
                same_action = game_object.getActionName(layer) == action_name
                if not same_action and is_playing:
                    game_object.stopAction(layer)
                if not (is_playing or same_action):
                    game_object.playAction(
                        action_name,
                        start_frame,
                        end_frame,
                        layer=layer,
                        priority=priority,
                        blendin=blendin,
                        play_mode=play_mode,
                        speed=speed,
                        layer_weight=1-layer_weight,
                        blend_mode=blend_mode)
                    self._started = True
                    self._frame = start_frame
                self._running = True
                self._finished = False
                self._finish_notified = False
            else:  # game_object is existing and valid but condition is False
                self._reset_subvalues()
        self.old_layer_weight = layer_weight
        self.old_speed = speed



from bge import logic
from uplogic.utils import STATUS_READY
from uplogic.utils import STATUS_WAITING
from uplogic.utils import check_game_object


def alpha_move(a, b, fac):
    if a < b:
        return a + fac
    elif a > b:
        return a - fac
    else:
        return a


_loaded_userlogic_files = {}


def load_user_logic(module_name):
    full_path = logic.expandPath(
        "//bgelogic/cells/{}.py".format(module_name)
    )
    loaded_value = _loaded_userlogic_files.get(full_path)
    if loaded_value:
        return loaded_value
    import sys
    python_version = sys.version_info
    major = python_version[0]
    minor = python_version[1]
    if (major < 3) or (major == 3 and minor < 3):
        import imp
        loaded_value = imp.load_source(module_name, full_path)
    elif (major == 3) and (minor < 5):
        from importlib.machinery import SourceFileLoader
        loaded_value = SourceFileLoader(module_name, full_path).load_module()
    else:
        import importlib.util
        spec = importlib.util.spec_from_file_location(module_name, full_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        loaded_value = module
    _loaded_userlogic_files[module_name] = loaded_value
    return loaded_value


class ULLogicBase(object):
    def get_value(self): pass
    def has_status(self, status): pass


class ULLogicContainer(ULLogicBase):

    def __init__(self):
        self._uid = None
        self._status = STATUS_WAITING
        self._value = None
        self._children = []
        self.network = None
        self.is_waiting = False

    def get_value(self):
        return self._value

    def _set_value(self, value):
        self._value = value

    def setup(self, network):
        """
        This is called by the network once, after all the
        cells have been loaded into the tree.
        :return: None
        """
        pass

    def stop(self, network):
        pass

    def _set_ready(self):
        self._status = STATUS_READY

    def _set_status(self, status):
        """
        Check the current status of the cell. Should return
        True if status equals the current status of the cell.
        :param status:
        :return:
        """
        self._status = status

    def has_status(self, status):
        return status == self._status

    def reset(self):
        """
        Resets the status of the cell to ULLogicContainer.STATUS_WAITING.
        A cell may override this to reset other states
        or to keep the value at STATUS_READY if evaluation is required
        to happen only once (or never at all)
        :return:
        """
        self._set_status(STATUS_WAITING)

    def evaluate(self):
        """
        A logic cell implements this method to do its job. The network
        evaluates a cell until its status becomes
         STATUS_READY. When that happens, the cell is
         removed from the update queue.
        :return:
        """
        raise NotImplementedError(
            "{} doesn't implement evaluate".format(self.__class__.__name__)
        )

    def _always_ready(self, status):
        return status is STATUS_READY

    def _skip_evaluate(self):
        return

    def deactivate(self):
        self.has_status = self._always_ready
        self.evaluate = self._skip_evaluate


###############################################################################
# Socket
###############################################################################


class ULOutSocket(ULLogicBase):

    def __init__(self, node, value_getter):
        self.node = node
        self.get_value = value_getter

    def has_status(self, status):
        return self.node.has_status(status)


###############################################################################
# Basic Cells
###############################################################################


class ULLogicNode(ULLogicContainer):

    def __init__(self):
        super().__init__()
        self.output_values = {}

    def reset(self):
        super().reset()
        self.output_values = {}

    def set_output(self, socket, value):
        self.output_values[socket] = value
        return value

    def get_output(self, socket, default=None):
        return self.output_values.get(socket, default)

    def get_input(self, param, scene=None):
        if str(param).startswith('NLO:'):
            if str(param) == 'NLO:U_O':
                return self.network._owner
            else:
                return check_game_object(param.split(':')[-1], scene)
        if isinstance(param, ULLogicBase):
            if param.has_status(STATUS_READY):
                return param.get_value()
            else:
                return STATUS_WAITING
        else:
            return param


class ULParameterNode(ULLogicNode):
    pass


class ULActionNode(ULLogicNode):
    pass


class ULConditionNode(ULLogicNode):
    pass



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import VEHICLE
from uplogic.utils import is_waiting
from uplogic.utils import is_invalid
from uplogic.utils import not_met
from uplogic.physics import FWD
from uplogic.physics import RWD
from uplogic.physics import FOURWD


class ULVehicleSetAttributes(ULActionNode):
    def __init__(self, value_type=RWD):
        ULActionNode.__init__(self)
        self.value_type = str(value_type)
        self.condition = None
        self.vehicle = None
        self.wheelcount = None
        self.set_suspension_compression = False
        self.suspension_compression = False
        self.set_suspension_stiffness = False
        self.suspension_stiffness = False
        self.set_suspension_damping = False
        self.suspension_damping = False
        self.set_tyre_friction = False
        self.tyre_friction = False
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def set_attributes(self, vehicle, wheel, attrs, values):
        if attrs[0] is True:
            vehicle.set_wheel_suspension(wheel, values[0])
        if attrs[1] is True:
            vehicle.set_wheel_stiffness(wheel, values[1])
        if attrs[2] is True:
            vehicle.set_wheel_damping(wheel, values[2])
        if attrs[3] is True:
            vehicle.set_wheel_friction(wheel, values[3])

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        game_object = self.get_input(self.vehicle)
        value_type = self.get_input(self.value_type)
        wheelcount = self.get_input(self.wheelcount)
        if is_waiting(value_type, wheelcount):
            return
        if is_invalid(game_object):
            print(game_object)
            return
        attrs_to_set = [
            self.get_input(self.set_suspension_compression),
            self.get_input(self.set_suspension_stiffness),
            self.get_input(self.set_suspension_damping),
            self.get_input(self.set_tyre_friction)
        ]
        values_to_set = [
            self.get_input(self.suspension_compression),
            self.get_input(self.suspension_stiffness),
            self.get_input(self.suspension_damping),
            self.get_input(self.tyre_friction)
        ]
        vehicle = game_object.get(VEHICLE, None)
        if not vehicle:
            return
        const = vehicle.constraint
        self._set_ready()
        if value_type == FWD:
            for wheel in range(wheelcount):
                self.set_attributes(
                    vehicle,
                    wheel,
                    attrs_to_set,
                    values_to_set
                )
        if value_type == RWD:
            for wheel in range(wheelcount):
                wheel = const.getNumWheels() - wheel - 1
                self.set_attributes(
                    vehicle,
                    wheel,
                    attrs_to_set,
                    values_to_set
                )
        if value_type == FOURWD:
            for wheel in range(const.getNumWheels()):
                self.set_attributes(
                    vehicle,
                    wheel,
                    attrs_to_set,
                    values_to_set
                )
        self.done = True



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULInstallSubNetwork(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.target_object = None
        self.tree_name = None
        self.initial_status = None
        self._network = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def setup(self, network):
        self._network = network

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        target_object = self.get_input(self.target_object)
        tree_name = self.get_input(self.tree_name)
        initial_status = self.get_input(self.initial_status)
        if is_waiting(
            target_object,
            tree_name,
            initial_status
        ):
            return
        self._set_ready()
        if is_invalid(target_object):
            return
        self._network.install_subnetwork(
            target_object,
            tree_name,
            initial_status
        )
        self.done = True



'''TODO: Documentation
'''

from types import MethodType
from bge import logic
import time


SCENE = logic.getCurrentScene()


def get_event_manager(name: str = 'default'):
    if not ULEventManager.initialized:
        SCENE.pre_draw.append(ULEventManager.update)
        ULEventManager.initialized = True


class ULEventManager():
    events = {}
    callbacks = []
    initialized = False
    done = []

    @classmethod
    def update(cls):
        for cb in cls.callbacks.copy():
            cb()

    @classmethod
    def log(cls):
        if cls.events:
            print('Events:')
            for evt in cls.events:
                print(f'\t{evt}:\t{cls.events[evt]}')

    @classmethod
    def schedule(cls, cb):
        if not cls.initialized:
            get_event_manager()
        cls.callbacks.append(cb)

    @classmethod
    def deschedule(cls, cb):
        if not cls.initialized:
            get_event_manager()
        cls.callbacks.remove(cb)

    @classmethod
    def register(cls, event):
        if not cls.initialized:
            get_event_manager()
        cls.events[event.name] = event
        cls.schedule(event.remove)

    @classmethod
    def catch(cls, name):
        if not cls.initialized:
            get_event_manager()
        return cls.events.get(name, None)

    @classmethod
    def consume(cls, name):
        if not cls.initialized:
            get_event_manager()
        return cls.events.pop(name, None)


class ULEvent():
    '''TODO: Documentation
    '''

    def __init__(self, name, content=None, messenger=None):
        self.name = name
        self.content = content
        self.messenger = messenger
        ULEventManager.schedule(self.register)

    def register(self):
        ULEventManager.register(self)
        ULEventManager.deschedule(self.register)

    def remove(self):
        ULEventManager.events.pop(self.name, None)
        ULEventManager.deschedule(self.remove)


def dispatch(name: str, content=None, messenger=None) -> None:
    '''TODO: Documentation
    '''
    ULEvent(name, content, messenger)


def handle(name: str):
    '''TODO: Documentation
    '''
    return ULEventManager.catch(name)


def consume(name: str):
    '''TODO: Documentation
    '''
    return ULEventManager.consume(name, None)


def schedule(name: str, content=None, messenger=None, delay=0.0):
    ScheduledEvent(delay, name, content, messenger)


class ScheduledEvent():

    def __init__(self, delay, name, content, messenger):
        self.time = time.time()
        self.delay = self.time + delay
        self.name = name
        self.content = content
        self.messenger = messenger
        ULEventManager.schedule(self.throw_scheduled)

    def throw_scheduled(self):
        if time.time() >= self.delay:
            ULEventManager.deschedule(self.throw_scheduled)
            ULEvent(self.name, self.content, self.messenger)


def schedule_callback(cb, delay=0.0, arg=None):
    ScheduledCallback(cb, delay, arg)


class ScheduledCallback():
    delay: float = 0
    arg = None
    callback: MethodType = None

    def __init__(self, cb, delay=0.0, arg=None):
        ULEventManager.schedule(self.call_scheduled)
        self.time = time.time()
        self.delay = self.time + delay
        self.callback = cb
        self.arg = arg

    def call_scheduled(self):
        if time.time() >= self.delay:
            if self.arg is not None:
                self.callback(self.arg)
            else:
                self.callback()
            ULEventManager.deschedule(self.call_scheduled)



from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import LOGIC_OPERATORS
from uplogic.utils import is_waiting


class ULCompare(ULConditionNode):
    def __init__(self, operator='GREATER'):
        ULConditionNode.__init__(self)
        self.operator = operator
        self.param_a = None
        self.param_b = None
        self.threshold = None
        self.RESULT = ULOutSocket(self, self.get_result)

    def get_result(self):
        socket = self.get_output('result')
        if socket is None:
            a = self.get_input(self.param_a)
            b = self.get_input(self.param_b)
            threshold = self.get_input(self.threshold)
            operator = self.get_input(self.operator)
            if is_waiting(a, b, threshold):
                return
            if operator > 1:  # eq and neq are valid for None
                if a is None:
                    return
                if b is None:
                    return
            if threshold is None:
                threshold = 0
            if threshold > 0 and abs(a - b) < threshold:
                a = b
            if operator is None:
                return
            return self.set_output(
                'result',
                LOGIC_OPERATORS[operator](a, b)
            )
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met
from uplogic.utils import debug


class ULSetGameObjectAttribue(ULActionNode):
    def __init__(self, value_type='worldPosition'):
        ULActionNode.__init__(self)
        self.value_type = str(value_type)
        self.condition = None
        self.xyz = None
        self.game_object = None
        self.attribute_value = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        xyz = self.get_input(self.xyz)
        game_object = self.get_input(self.game_object)
        attribute = self.get_input(self.value_type)
        value = self.get_input(self.attribute_value)
        if is_waiting(xyz, game_object, attribute, value):
            return

        if hasattr(value, attribute):
            value = getattr(value, attribute).copy()
        self._set_ready()
        if not hasattr(game_object, attribute):
            debug(
                'Set Object Data Node: {} has no attribute {}!'
                .format(game_object, attribute)
            )
            return
        data = getattr(game_object, attribute)
        if 'Orientation' in attribute:
            data = data.to_euler()
        for axis in xyz:
            if not xyz[axis]:
                setattr(value, axis, getattr(data, axis))
        setattr(
            game_object,
            attribute,
            value
        )
        if value == 'worldScale':
            game_object.reinstancePhysicsMesh(
                game_object,
                game_object.meshes[0]
            )
        self.done = True



from bge import logic
from random import randint, random
from bge.types import KX_GameObject as GameObject
from uplogic.animation import ULActionSystem
from uplogic.data import GlobalDB


PLAY_MODES = {
    'play': logic.KX_ACTION_MODE_PLAY,
    'pingpong': logic.KX_ACTION_MODE_PING_PONG,
    'loop': logic.KX_ACTION_MODE_LOOP
}


BLEND_MODES = {
    'blend': logic.KX_ACTION_BLEND_BLEND,
    'add': logic.KX_ACTION_BLEND_ADD
}


class ULAction():
    '''TODO: Documentation
    '''
    game_object = None
    act_system = 'default'
    acion_name: str
    start_frame: int
    end_frame: int
    layer: int
    priority: int
    blendin: float
    play_mode: str
    blend_mode: int
    speed: float
    old_speed: float = 0
    layer_weight: float
    old_layer_weight: float = 0
    frame: float = 0

    def __init__(
        self,
        game_object: GameObject,
        action_name: str,
        start_frame: int = 0,
        end_frame: int = 250,
        layer: int = -1,
        priority: int = 0,
        blendin: float = 0,
        play_mode: str = 'play',
        speed: float = 1,
        layer_weight: float = 1,
        blend_mode: str = 'blend',
    ):
        self.act_system = self.get_act_sys(self.act_system)
        self.layer = layer
        self.game_object = game_object
        if self.layer == -1:
            ULActionSystem.find_free_layer(self)
        elif ULActionSystem.check_layer(self):
            return
        layer = self.layer
        self.acion_name = action_name
        self.start_frame = start_frame
        self.end_frame = end_frame
        self.priority = priority
        self.blendin = blendin
        self.play_mode = play_mode
        self.blend_mode = blend_mode
        self.speed = speed
        self.layer_weight = layer_weight
        is_playing = game_object.isPlayingAction(layer)
        same_action = game_object.getActionName(layer) == action_name
        if not same_action and is_playing:
            game_object.stopAction(layer)
        if not (is_playing or same_action):
            game_object.playAction(
                action_name,
                start_frame,
                end_frame,
                play_mode=PLAY_MODES[play_mode],
                speed=speed,
                layer=layer,
                priority=priority,
                blendin=blendin,
                layer_weight=1-layer_weight,
                blend_mode=BLEND_MODES[blend_mode])
            self._started = True
            self._frame = start_frame
        self.act_system.add(self)

    def update(self):
        layer_weight = self.layer_weight
        speed = self.speed
        if layer_weight <= 0:
            layer_weight = 0.0
        elif layer_weight >= 1:
            layer_weight = 1.0
        if speed <= 0:
            speed = 0.01
        game_object = self.game_object
        layer = self.layer
        start_frame = self.start_frame
        end_frame = self.end_frame
        action_name = self.acion_name
        play_mode = self.play_mode
        priority = self.priority
        blendin = self.blendin
        blend_mode = self.blend_mode
        playing_action = game_object.getActionName(layer)
        playing_frame = game_object.getActionFrame(layer)
        min_frame = start_frame
        max_frame = end_frame
        if end_frame < start_frame:
            min_frame = end_frame
            max_frame = max_frame
        if (
            (playing_action == action_name) and
            (playing_frame >= min_frame) and
            (playing_frame <= max_frame)
        ):
            self._frame = playing_frame
            if (
                layer_weight != self.old_layer_weight or
                speed != self.old_speed
            ):
                reset_frame = (
                    start_frame if
                    play_mode == logic.KX_ACTION_MODE_LOOP else
                    end_frame
                )
                next_frame = (
                    playing_frame + speed
                    if
                    playing_frame + speed <= end_frame
                    else
                    reset_frame
                )
                game_object.stopAction(layer)
                game_object.playAction(
                    action_name,
                    start_frame,
                    end_frame,
                    layer=layer,
                    priority=priority,
                    blendin=blendin,
                    play_mode=PLAY_MODES[play_mode],
                    speed=speed,
                    layer_weight=1 - layer_weight,
                    blend_mode=BLEND_MODES[blend_mode]
                )
                game_object.setActionFrame(next_frame, layer)
            if play_mode == 'play':
                if end_frame > start_frame:  # play 0 to 100
                    is_near_end = (playing_frame >= (end_frame - 0.5))
                else:  # play 100 to 0
                    is_near_end = (playing_frame <= (end_frame + 0.5))
                if is_near_end:
                    self.act_system.remove(self)
        self.frame = game_object.getActionFrame(layer)
        self.old_layer_weight = layer_weight
        self.old_speed = speed

    def stop(self):
        self.game_object.stopAction(self.layer)

    def randomize_frame(self, min=None, max=None):
        if min is None:
            min = self.start_frame
        if max is None:
            max = self.end_frame
        frame = randint(min, max)
        self.frame = frame
        self.game_object.setActionFrame(
            frame,
            self.layer
        )

    def randomize_speed(self, min=.9, max=1.1):
        delta = max - min
        self.speed = min + (delta * random())

    def set_frame(self, frame):
        self.frame = frame
        self.game_object.setActionFrame(
            frame,
            self.layer
        )

    def get_act_sys(self, name: str):
        act_systems = GlobalDB.retrieve('uplogic.animation')
        if act_systems.check(name):
            return act_systems.get(name)
        else:
            return ULActionSystem(name)



from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
import bpy


class ULGetImage(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.image = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        image = self.get_input(self.image)
        if is_invalid(image):
            return
        return bpy.data.images[image]

    def evaluate(self):
        self._set_ready()



from .actionstatus import ULActionStatus  # noqa
from .activecamera import ULActiveCamera  # noqa
from .axisvector import ULAxisVector  # noqa
from .bonestatus import ULBoneStatus  # noqa
from .characterinfo import ULCharacterInfo  # noqa
from .childbyindex import ULChildByIndex  # noqa
from .childbyname import ULChildByName  # noqa
from .clamp import ULClamp  # noqa
from .colorrgb import ULColorRGB  # noqa
from .colorrgba import ULColorRGBA  # noqa
from .dictvalue import ULDictValue  # noqa
from .distance import ULDistance  # noqa
from .euler import ULEuler  # noqa
from .eulertomatrix import ULEulerToMatrix  # noqa
from .formattedstring import ULFormattedString  # noqa
from .formula import ULFormula  # noqa
from .gamepadsticks import ULGamepadSticks  # noqa
from .gamepadtrigger import ULGamepadTrigger  # noqa
from .getactuatorvalue import ULGetActuatorValue  # noqa
from .getcollection import ULGetCollection  # noqa
from .getcollectionobjectnames import ULGetCollectionObjectNames  # noqa
from .getcollectionobjects import ULGetCollectionObjects  # noqa
from .getcurvepoints import ULGetCurvePoints  # noqa
from .getglobalvalue import ULGetGlobalValue  # noqa
from .getgravity import ULGetGravity  # noqa
from .getimage import ULGetImage  # noqa
from .getnodeattribute import ULGetNodeAttribute  # noqa
from .getnodesocket import ULGetNodeSocket  # noqa
from .getobject import ULGetObject  # noqa
from .getowner import ULGetOwner  # noqa
from .getparent import ULGetParent  # noqa
from .getproperty import ULGetProperty  # noqa
from .getpyinstanceattr import ULGetPyInstanceAttr  # noqa
from .getscene import ULGetScene  # noqa
from .getsensorvalue import ULGetSensorValue  # noqa
from .getsound import ULGetSound  # noqa
from .gettimescale import ULGetTimeScale  # noqa
from .interpolate import ULInterpolate  # noqa
from .keycode import ULKeyCode  # noqa
from .limitrange import ULLimitRange  # noqa
from .listduplicate import ULListDuplicate  # noqa
from .listfromitems import ULListFromItems  # noqa
from .listindex import ULListIndex  # noqa
from .listindexrandom import ULListIndexRandom  # noqa
from .materialgetattribute import ULGetMaterialAttribute  # noqa
from .materialgetnode import ULGetMaterialNode  # noqa
from .materialgetsocket import ULGetMaterialSocket  # noqa
from .math import ULMath  # noqa
from .matrixtoxyz import ULMatrixToXYZ  # noqa
from .mousedata import ULMouseData  # noqa
from .objectattr import ULObjectAttribute  # noqa
from .objectdataname import ULObjectDataName  # noqa
from .objectdatavertices import ULObjectDataVertices  # noqa
from .randomfloat import ULRandomFloat  # noqa
from .rangedthreshold import ULRangedThreshold  # noqa
from .screenposition import ULScreenPosition  # noqa
from .simplevalue import ULSimpleValue  # noqa
from .threshold import ULThreshold  # noqa
from .timedata import ULTimeData  # noqa
from .typecastvalue import ULTypeCastValue  # noqa
from .vectorabsolute import ULVectorAbsolute  # noqa
from .vectorangle import ULVectorAngle  # noqa
from .vectoranglecheck import ULVectorAngleCheck  # noqa
from .vectorlength import ULVectorLength  # noqa
from .vectormath import ULVectorMath  # noqa
from .vectorsplitxy import ULVectorSplitXY  # noqa
from .vectorsplitxyz import ULVectorSplitXYZ  # noqa
from .vectorxy import ULVectorXY  # noqa
from .vectorxyz import ULVectorXYZ  # noqa
from .vectorxyzw import ULVectorXYZW  # noqa
from .withinrange import ULWithinRange  # noqa
from .worldposition import ULWorldPosition  # noqa

from .getvsync import ULGetVSync  # noqa
from .invertvalue import ULInvertValue  # noqa
from .valueswitch import ULValueSwitch  # noqa
from .getfullscreen import ULGetFullscreen  # noqa
from .getresolution import ULGetResolution  # noqa
from .initemptydict import ULInitEmptyDict  # noqa
from .initnewdict import ULInitNewDict  # noqa
from .initemptylist import ULInitEmptyList  # noqa
from .loadvariable import ULLoadVariable
from .loadvariabledict import ULLoadVariableDict
from .randomint import ULRandomInt
from .randomvect import ULRandomVect
from .getlightenergy import ULGetLightEnergy
from .getlightcolor import ULGetLightColor



from bge import render
from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULRaycast(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.origin = None
        self.destination = None
        self.local: bool = None
        self.property_name: str = None
        self.xray: bool = None
        self.custom_dist: bool = None
        self.distance: float = None
        self.visualize: bool = None
        self._picked_object = None
        self._point = None
        self._normal = None
        self._direction = None
        self.PICKED_OBJECT = ULOutSocket(self, self.get_picked_object)
        self.POINT = ULOutSocket(self, self.get_point)
        self.NORMAL = ULOutSocket(self, self.get_normal)
        self.DIRECTION = ULOutSocket(self, self.get_direction)
        self.network = None

    def setup(self, network):
        self.network = network

    def get_picked_object(self):
        return self._picked_object

    def get_point(self):
        return self._point

    def get_normal(self):
        return self._normal

    def get_direction(self):
        return self._direction

    def _compute_direction(self, origin, dest, local, dist):
        custom_dist = self.get_input(self.custom_dist)
        start = (
            origin.worldPosition.copy()
            if hasattr(origin, "worldPosition")
            else origin
        )
        if hasattr(dest, "worldPosition"):
            dest = dest.worldPosition.copy()
        if local:
            dest = start + dest
        d = dest - start
        d.normalize()
        return d, dist if custom_dist else (start - dest).length, dest

    def evaluate(self):
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_value(False)
            self._normal = None
            self._object = None
            return
        origin = self.get_input(self.origin)
        destination = self.get_input(self.destination)
        local: bool = self.get_input(self.local)
        property_name: str = self.get_input(self.property_name)
        xray: bool = self.get_input(self.xray)
        distance: float = self.get_input(self.distance)
        visualize: bool = self.get_input(self.visualize)

        if is_waiting(origin, destination, local, property_name, distance):
            return
        self._set_ready()
        caster = self.network._owner
        obj, point, normal = None, None, None
        direction, distance, destination = self._compute_direction(
            origin,
            destination,
            local,
            distance
        )
        if not property_name:
            obj, point, normal = caster.rayCast(
                destination,
                origin,
                distance,
                xray=xray
            )
        else:
            obj, point, normal = caster.rayCast(
                destination,
                origin,
                distance,
                property_name,
                xray=xray
            )
        if visualize:
            origin = getattr(origin, 'worldPosition', origin)
            line_dest: Vector = direction.copy()
            line_dest.x *= distance
            line_dest.y *= distance
            line_dest.z *= distance
            line_dest = line_dest + origin
            render.drawLine(
                origin,
                line_dest,
                [
                    1,
                    0,
                    0,
                    1
                ]
            )
            if obj:
                render.drawLine(
                    origin,
                    point,
                    [
                        0,
                        1,
                        0,
                        1
                    ]
                )
        self._set_value(obj is not None)
        self._picked_object = obj
        self._point = point
        self._normal = normal
        self._direction = direction



from uplogic.events import handle
from uplogic.events import ULEvent
from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid


class ULHandleEvent(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.subject = None
        self.received = False
        self.event = None
        self.OUT = ULOutSocket(self, self.get_received)
        self.BODY = ULOutSocket(self, self.get_body)
        self.TARGET = ULOutSocket(self, self.get_target)

    def get_received(self):
        return isinstance(self.event, ULEvent)

    def get_body(self):
        return STATUS_WAITING if self.event is None else self.event.content

    def get_target(self):
        return STATUS_WAITING if self.event is None else self.event.messenger

    def evaluate(self):
        subject = self.get_input(self.subject)
        if is_invalid(subject):
            return
        self._set_ready()
        self.event = handle(subject)



from bge import logic
from mathutils import Vector
from uplogic.utils import vec_clamp
from uplogic.utils import raycast
from uplogic.utils import FLOATSAM
from uplogic.utils import SHIP
from uplogic.utils import WATER


class ULBuoy():

    def __init__(self) -> None:
        self._active = True

    def disable(self):
        self._active = False

    def enable(self):
        self._active = True

    def destroy(self):
        raise NotImplementedError


class ULFloatsam(ULBuoy):

    def __init__(self, game_object, buoyancy=1, height=200, align=True) -> None:
        super().__init__()
        self.game_object = game_object
        game_object[FLOATSAM] = self
        self.height = height
        self.buoyancy = buoyancy
        self.align = align
        logic.getCurrentScene().pre_draw.append(self.update)

    def update(self):
        if not self._active:
            return
        up = Vector((0, 0, 1))
        floatsam = self.game_object
        lindamp = .1
        wpos = floatsam.worldPosition
        obj, point, normal, direction = raycast(
            floatsam,
            wpos,
            up,
            self.height,
            WATER,
            True,
            True
        )
        if obj:
            lindamp = .8
            lift = (up * (wpos - point).length * self.buoyancy)
            floatsam.applyImpulse(
                wpos,
                vec_clamp(lift, max=self.buoyancy),
                False
            )
            if self.align:
                self.game_object.alignAxisToVect(normal, 2, .2)
        floatsam.linearDamping = lindamp
        floatsam.angularDamping = lindamp * .8

    def destroy(self):
        logic.getCurrentScene().pre_draw.remove(self.update)


class ULShip(ULBuoy):

    def __init__(self, game_object, buoyancy=1, height=200) -> None:
        super().__init__()
        self.game_object = game_object
        game_object[SHIP] = self
        print(game_object)
        self.height = height
        cs = sorted(game_object.childrenRecursive, key=lambda c: c.name)
        self.buoys = [c for c in cs if 'Buoy' in c.name]
        self.buoyancy = buoyancy
        logic.getCurrentScene().pre_draw.append(self.update)

    def update(self):
        if not self._active:
            return
        up = Vector((0, 0, 1))
        lifts = len(self.buoys)
        ship = self.game_object
        lindamp = .1
        for buoy in self.buoys:
            wpos = buoy.worldPosition
            obj, point, normal, direction = raycast(
                buoy,
                wpos,
                up,
                self.height,
                WATER,
                True,
                True,
                True
            )
            if obj:
                lindamp += (.7 / lifts)
                lift = (up * (wpos - point).length * self.buoyancy) / lifts
                ship.applyImpulse(
                    wpos,
                    vec_clamp(lift, max=self.buoyancy * 2 / lifts),
                    False
                )
        ship.linearDamping = lindamp
        ship.angularDamping = lindamp * .8

    def destroy(self):
        logic.getCurrentScene().pre_draw.remove(self.update)



from bge.types import KX_GameObject
from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULLogicContainer
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting


class ULCollision(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.game_object = None
        self.use_mat = None
        self.prop = None
        self.material = None
        self._set_value("False")
        self.pulse = False
        self._target = None
        self._point = None
        self._normal = None
        self._collision_triggered = False
        self._consumed = False
        self._last_monitored_object = None
        self._objects = []
        self.TARGET = ULOutSocket(self, self.get_target)
        self.POINT = ULOutSocket(self, self.get_point)
        self.NORMAL = ULOutSocket(self, self.get_normal)
        self.OBJECTS = ULOutSocket(self, self.get_objects)

    def get_point(self):
        return self._point

    def get_normal(self):
        return self._normal

    def get_target(self):
        return self._target

    def get_objects(self):
        return self._objects

    def _collision_callback(self, obj, point, normal):
        self._objects.append(obj)
        use_mat = self.get_input(self.use_mat)
        if use_mat:
            material = self.get_input(self.material)
            if material:
                for obj in self._objects:
                    bo = obj.blenderObject
                    if material not in [
                        slot.material.name for
                        slot in
                        bo.material_slots
                    ]:
                        self._objects.remove(obj)
                    else:
                        self._collision_triggered = True
                        self._target = obj
                        self._point = point
                        self._normal = normal
                        return
                self._collision_triggered = False
                return
        else:
            prop = self.get_input(self.prop)
            if prop:
                for obj in self._objects:
                    if prop not in obj:
                        self._objects.remove(obj)
                    else:
                        self._collision_triggered = True
                        self._target = obj
                        self._point = point
                        self._normal = normal
                        return
                self._collision_triggered = False
                return
        self._collision_triggered = True
        self._target = obj
        self._point = point
        self._normal = normal

    def reset(self):
        ULLogicContainer.reset(self)
        self._collision_triggered = False
        self._objects = []

    def _reset_last_monitored_object(self, new_monitored_object):
        if is_invalid(new_monitored_object):
            new_monitored_object = None
        if self._last_monitored_object == new_monitored_object:
            return
        if not isinstance(new_monitored_object, KX_GameObject):
            if self._last_monitored_object is not None:
                self._last_monitored_object.collisionCallbacks.remove(
                    self._collision_callback
                )
                self._last_monitored_object = None
        else:
            if self._last_monitored_object is not None:
                self._last_monitored_object.collisionCallbacks.remove(
                    self._collision_callback
                )
            if new_monitored_object is not None:
                new_monitored_object.collisionCallbacks.append(
                    self._collision_callback
                )
                self._last_monitored_object = new_monitored_object
        self._set_value(False)
        self._target = None
        self._point = None
        self._normal = None
        self._collision_triggered = False

    def evaluate(self):
        last_target = self._target
        game_object = self.get_input(self.game_object)
        self._reset_last_monitored_object(game_object)
        if is_waiting(game_object):
            return
        self._set_ready()
        collision = self._collision_triggered
        if last_target is not self._target:
            self._consumed = False
        if collision and not self.pulse:
            self._set_value(collision and not self._consumed)
            self._consumed = True
        elif self.pulse:
            self._set_value(collision)
        else:
            self._consumed = False
            self._set_value(False)



from bge import logic
from uplogic.nodes import ULActionNode
from uplogic.utils import is_invalid
from uplogic.utils import not_met
import bpy


class ULSetOverlayCollection(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.camera = None
        self.collection = None

    def evaluate(self):
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        collection = self.get_input(self.collection)
        camera = self.get_input(self.camera)
        if is_invalid(camera, collection):
            return
        self._set_ready()
        col = bpy.data.collections.get(collection)
        if not col:
            return
        logic.getCurrentScene().addOverlayCollection(camera, col)
        self._set_value(True)



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting


class ULActionStatus(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.action_layer = None
        self._action_name = None
        self._action_frame = None
        self.NOT_PLAYING = ULOutSocket(self, self.get_not_playing)
        self.ACTION_NAME = ULOutSocket(self, self.get_action_name)
        self.ACTION_FRAME = ULOutSocket(self, self.get_action_frame)

    def get_action_name(self):
        return self._action_name

    def get_action_frame(self):
        return self._action_frame

    def get_not_playing(self):
        return not self.get_value()

    def evaluate(self):
        game_object = self.get_input(self.game_object)
        action_layer = self.get_input(self.action_layer)
        if is_waiting(game_object, action_layer):
            return
        self._set_ready()
        if is_invalid(game_object):
            self._action_name = STATUS_WAITING
            self._action_frame = STATUS_WAITING
            self._set_value(False)
        else:
            self._set_value(game_object.isPlayingAction(action_layer))
            self._action_name = game_object.getActionName(action_layer)
            self._action_frame = game_object.getActionFrame(action_layer)



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met
import bpy


class ULSetActionFrame(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.action_layer = None
        self.action_frame = None
        self.freeze = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        game_object = self.get_input(self.game_object)
        action_layer = self.get_input(self.action_layer)
        action_frame = self.get_input(self.action_frame)
        freeze = self.get_input(self.freeze)
        action_name = self.get_input(self.action_name)
        layer_weight = self.get_input(self.layer_weight)
        self._set_ready()
        if is_waiting(
            action_layer,
            action_frame,
            layer_weight
        ):
            return
        if is_invalid(
            game_object,
        ):
            return
        is_playing = game_object.isPlayingAction(action_layer)
        same_action = game_object.getActionName(action_layer) == action_name
        action = bpy.data.actions[action_name]
        start_frame = action.frame_range[0]
        end_frame = action.frame_range[1]
        finished = game_object.getActionFrame(action_layer) >= end_frame
        if not (is_playing or same_action) or finished:
            game_object.stopAction(action_layer)
            speed = .000000000000000001 if freeze else 1
            game_object.playAction(
                action_name,
                start_frame,
                end_frame,
                action_layer,
                layer_weight=1-layer_weight,
                speed=speed
            )
        game_object.setActionFrame(action_frame, action_layer)
        self.done = True



from bge import logic
from bge import events


KEYBOARD_EVENTS = logic.keyboard.inputs
'''Reference to `bge.logic.keyboard.inputs`
'''

_keys_active = {}


def key_event(key: str) -> bool:
    '''Retrieve key event.\n
    Not intended for manual use.
    '''
    return KEYBOARD_EVENTS[
        getattr(
            events, f'{key}KEY',
            (getattr(events, f'PAD{key}', None))
        )
    ]


def pad_event(key: str) -> bool:
    '''Retrieve Numpad event.\n
    Not intended for manual use.
    '''
    return KEYBOARD_EVENTS[getattr(events, f'PAD{key}')]


def key_tap(key: str) -> bool:
    '''Detect key tapped.

    :param key: key as `str` (e.g. `'A'`)

    :returns: boolean
    '''
    return key_event(key).activated


def key_down(key: str) -> bool:
    '''Detect key held down.

    :param key: key as `str` (e.g. `'A'`)

    :returns: boolean
    '''
    return key_event(key).active


def key_up(key: str) -> bool:
    '''Detect key released.

    :param key: key as `str` (e.g. `'A'`)

    :returns: boolean
    '''
    return key_event(key).released


def key_pulse(key: str, time: float = .4) -> bool:
    '''Detect key tapped, then held down after `time` has passed.

    :param key: key as `str` (e.g. `'A'`)
    :param time: timeout for key down

    :returns: boolean
    '''
    if key_event(key).activated:
        _keys_active[key] = 0
        return True
    k = _keys_active.get(key, 0)
    _keys_active[key] = k + (1 / logic.getAverageFrameRate())
    if _keys_active[key] > time:
        return key_event(key).active
    return False
